/*
 * Copyright (C) 2003-2004 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//
// @author Stefano Fornari
// @version $Id: SyncMLBuilder.cpp,v 1.10 2004/04/15 16:44:27 magi Exp $
//

#include <stdlib.h>

#include "common/Log.h"
#include "spds/common/Constants.h"
#include "spds/common/SyncMLBuilder.h"

#include "spds/common/MessageTemplates.h"

USING_PTYPES

/*
 * Constructors
 */

SyncMLBuilder::SyncMLBuilder() {
    wcscpy(target,   TEXT(""));
    wcscpy(username, TEXT(""));
    wcscpy(password, TEXT(""));
    wcscpy(device,   TEXT(""));
}

SyncMLBuilder::SyncMLBuilder(wchar_t*      t  ,
                             wchar_t*      u,
                             wchar_t*      p,
                             wchar_t*      d) {
    wcsncpy(target,   t, DIM_SERVERNAME); target  [DIM_SERVERNAME-1] = 0;
    wcsncpy(username, u, DIM_USERNAME  ); username[DIM_USERNAME-1  ] = 0;
    wcsncpy(password, p, DIM_PASSWORD  ); password[DIM_PASSWORD-1  ] = 0;
    wcsncpy(device,   d, DIM_DEVICE_ID ); device  [DIM_DEVICE_ID-1 ] = 0;
}

/*
 * Creates and returns the initialization message for a given source.
 * The returned pointer points to a memory buffer allocated with
 * the new operator and must be discarded with the delete oprator.
 *
 * @param source - The SyncSource - NOT NULL
 *
 */
wchar_t* SyncMLBuilder::prepareInitMsg(SyncSource& source) {
    unsigned int size;
    string alert = string();

    createAlert(source, alert);


    //
    // Allocate anough memory for the entire message (alert
    // + xmlInit + info)
    //
    size = wcslen(xmlInit)
         + length(alert)
         + 1024              // target + deviceid + username + password + source
         ;
    wchar_t* buf = new wchar_t[size];
    *buf = 0;

    wsprintf(buf                     ,
             xmlInit                 ,
             target                  ,
             device                  ,
             username                ,
             password                ,
             (const wchar_t*)alert   );

    return buf;
}

/*
 * Creates and returns the synchronization message for the given source.
 * The returned pointer points to a memory buffer allocated with
 * the new operator and must be discarded with the delete oprator.
 *
 * @param source the SyncSource to synchronize
 */
wchar_t* SyncMLBuilder::prepareSyncMsg(SyncSource& source) {
    wchar_t logmsg[512];

    SyncItem** items = NULL;
    unsigned int n, i;
    string statusTag = string();
    string syncTag = string();

    switch (source.getSyncMode()) {
        case SYNC_SLOW:
            //
            // Slow Sync!
            //
            wsprintf(logmsg, TEXT("Preparing slow sync for %s"), source.getName(NULL, 0));
            LOG.debug(logmsg);

            n = source.getAllSyncItemsCount();

            items = source.getAllSyncItems();

            break;

        case SYNC_REFRESH_FROM_SERVER:
            //
            // Refresh from server
            //
            wsprintf(logmsg, TEXT("Prepare refresh for %s"), source.getName(NULL, 0));
            LOG.debug(logmsg);
            items = new SyncItem*[0];

            break;

        case SYNC_ONE_WAY_FROM_SERVER:
            //
            // One-way sync from server: no needs of client modifications
            //
            wsprintf(logmsg, TEXT("Prepare one-way sync for %s"), source.getName(NULL, 0));
            LOG.debug(logmsg);
            items = new SyncItem*[0];

            break;

        default:
            //
            // Fast Sync!
            //
            wsprintf(logmsg, TEXT("Prepare fast sync for %s"), source.getName(NULL, 0));
            LOG.debug(logmsg);

            //
            // NOTE: filterRecordsForFastSync returns items in updated state
            //
            n = source.getNewSyncItemsCount()
              + source.getUpdatedSyncItemsCount()
              + source.getDeletedSyncItemsCount()
              ;

            items = new SyncItem*[n];

            i = source.getNewSyncItemsCount();
            memmove(items, source.getNewSyncItems(), i*sizeof(SyncItem*));
            n = i;
            i = source.getUpdatedSyncItemsCount();
            memmove(&items[n], source.getUpdatedSyncItems(), i*sizeof(SyncItem*));
            n += i;
            i = source.getDeletedSyncItemsCount();
            memmove(&items[n], source.getDeletedSyncItems(), i*sizeof(SyncItem*));
            n += i;

            break;
    }

    createSync(source, items, n, syncTag);

    //
    // Calculate the size for the entire message
    //
    unsigned int size = wcslen(xmlSync1)
                      + wcslen(xmlSync2)
                      + length(syncTag)
                      + 1024            // target + device + 2*source + timestamp
                      ;

    wchar_t* buf = new wchar_t[size];
    wsprintf(buf                       ,
             xmlSync1                  ,
             target                    ,
             device                    ,
             target                    ,
             device                    ,
             source.getName(NULL, 0)   ,
             source.getName(NULL, 0)   ,
             source.getNextAnchor(NULL),
             source.getName(NULL, 0)   ,
             source.getName(NULL, 0)   );
    wcscat(buf, (wchar_t*)syncTag);
    wcscat(buf, xmlSync2);

    return buf;
}

/*
 * Creates and returns the mapping message for a given source.
 * The returned pointer points to a memory buffer allocated with
 * the new operator and must be discarded with the delete oprator.
 *
 * @param source - The SyncSource - NOT NULL
 *
 */
wchar_t* SyncMLBuilder::prepareMapMsg(SyncSource& source) {
    string map = string();

    createMap(source, map);


    //
    // Calculate the size for the entire message
    //
    unsigned int size = wcslen(xmlMap)
                      + length(map)
                      + 1024            // target + device
                      ;

    wchar_t* buf = new wchar_t[size];
/*
    wsprintf( buf                 ,
              xmlMap              ,
              target              ,
              device              ,
              target              ,
              device              ,
              (const wchar_t*)map );
*/

	wsprintf( buf                 ,
              xmlMapStart1        ,
              target              ,
              device              ,
              target              ,
              device              );
	
	// work fine ***
	//wchar_t* tmp = new wchar_t[size];	
	//wsprintf(tmp, TEXT("%s%s"),(wchar_t*)map, xmlMapStart2);	
	//wcscat(buf, tmp);
	//delete tmp;
	//***
	
	//work fine
	//wcscat(buf, (wchar_t*)map);
	//wcscat(buf, xmlMapStart2);
	//****

	concat(buf, (wchar_t*)map);

	concat(buf, xmlMapStart2);


    return buf;
}


// ----------------------------------------------------------------- Private methods

void SyncMLBuilder::createAlert(SyncSource& source, string& alert) {
    wchar_t buf[32];

    wsprintf(buf, TEXT("%d"), source.getPreferredSyncMode());

    alert = TEXT("<Alert>\n<CmdID>1</CmdID>\n<Data>");
    alert += buf;
    alert += TEXT("</Data>\n<Item>\n<Target><LocURI>");
    alert += source.getName(NULL, 0);
    alert += TEXT("</LocURI></Target>\n<Source><LocURI>");
    alert += source.getName(NULL, 0);
    alert += TEXT("</LocURI></Source>\n<Meta><Anchor xmlns=\"syncml:metinf\">\n");
    alert += TEXT("<Last>");
    alert += source.getLastAnchor(NULL);
    alert += TEXT("</Last>\n");
    alert += TEXT("<Next>");
    alert += source.getNextAnchor(NULL);
    alert += TEXT("</Next>\n");
    alert += TEXT("</Anchor></Meta>\n</Item>\n</Alert>\n");
}

void SyncMLBuilder::createSync(SyncSource& source, SyncItem* items[], int n, string& syncTag) {
    int i = 0;

    string addItems     = string();
    string deleteItems  = string();
    string replaceItems = string();

    SyncMode syncMode = source.getSyncMode();

    for(i=0; i < n; ++i) {
        switch (items[i]->getState()) {
            case SYNC_STATE_DELETED:
                deleteItems += TEXT("<Item>\n<Source><LocURI>");
                deleteItems += items[i]->getKey(NULL);
                deleteItems += TEXT("</LocURI></Source>\n</Item>\n");
                break;

            case SYNC_STATE_UPDATED:
            case SYNC_STATE_NONE:  // this is for slow sync
                replaceItems += TEXT("<Item>\n<Source><LocURI>");
                replaceItems += items[i]->getKey(NULL);
                replaceItems += TEXT("</LocURI></Source>\n<Data>\n");
                replaceItems += (wchar_t*)items[i]->getData();
                replaceItems += TEXT("</Data>\n</Item>\n");
                break;

            case SYNC_STATE_NEW:
                addItems += TEXT("<Item>\n<Source><LocURI>");
                addItems += items[i]->getKey(NULL);
                addItems += TEXT("</LocURI></Source>\n<Data>\n");
                addItems += (wchar_t*)items[i]->getData();
                addItems += TEXT("</Data>\n</Item>\n");
                break;
        } // end switch
    }

    wchar_t cmdID = '4';
    if (length(addItems)>0) {
        syncTag += TEXT("\n<Add>\n<CmdID>");
        syncTag += cmdID;
        syncTag += TEXT("</CmdID>\n<Meta><Type xmlns='syncml:metinf'>");
        syncTag += source.getType(NULL);
        syncTag += TEXT("</Type></Meta>\n");
        syncTag += addItems;
        syncTag += TEXT("</Add>");

        ++cmdID;
    }

    if (length(replaceItems)>0) {
        syncTag += TEXT("\n<Replace>\n<CmdID>");
        syncTag += cmdID;
        syncTag += TEXT("</CmdID>\n<Meta><Type xmlns='syncml:metinf'>");
        syncTag += source.getType(NULL);
        syncTag += TEXT("</Type></Meta>\n");
        syncTag += replaceItems;
        syncTag += TEXT("</Replace>");

        ++cmdID;
    }

    if (length(deleteItems)>0) {
        syncTag += TEXT("\n<Delete>\n<CmdID>");
        syncTag += cmdID;
        syncTag += TEXT("</CmdID>\n");
        syncTag += deleteItems;
        syncTag += TEXT("</Delete>");

        ++cmdID;
    }
}

void SyncMLBuilder::createMap(SyncSource& source, string& map) {
    unsigned int n;

    //
    // the MapItem element must be inserted only if there are
    // mappings
    //
    n = source.getMapSize();
    if (n == 0) {
        return;
    }

    map += TEXT("<Map>\n<CmdID>3</CmdID>\n<Target><LocURI>");
    map += source.getName(NULL, 0);
    map += TEXT("</LocURI></Target>\n<Source><LocURI>");
    map += source.getName(NULL, 0);
    map += TEXT("</LocURI></Source>\n");

    SyncMap** mappings = source.getLUIDGUIDMapping();

    for (unsigned int i=0; i<n; ++i) {
        map += TEXT("<MapItem>\n<Target><LocURI>");
        map += mappings[i]->getGUID(NULL);
        map += TEXT("</LocURI></Target>\n");
        map += TEXT("<Source><LocURI>");
        map += mappings[i]->getLUID(NULL);
        map += TEXT("</LocURI></Source>\n</MapItem>\n");
    }

    map += TEXT("</Map>");
}

/* 
* For debugging the concat from buf with string...
*/

void SyncMLBuilder::concat(wchar_t* buf, wchar_t* map) {
    unsigned int n, bufLen;
	
    
	n = wcslen(map);
	bufLen = wcslen(buf);

    if (n == 0) {
        return;
    }		
	
    for (unsigned int i=0; i<n; ++i) {		
		if (map[i] != 0)
			buf[bufLen + i] = map[i];        
		else {
			wchar_t* localTemp = new wchar_t[bufLen + i + 256];
			wsprintf(localTemp, TEXT("SC-API: %s - 0 at char num %i"), buf, i);
			MessageBox (NULL, localTemp, TEXT ("Debug eVC: Found 0"), MB_SETFOREGROUND |MB_OK);
			delete localTemp;
			break;
		}

    }
	
	buf[bufLen + n] = 0;


}	
